#include "Icon.h"
#include <iostream>

  Icon::Icon( Desk *Dsk, char *File, char *F, char *Cap, char *Com, int X, int Y ) {
    XSetWindowAttributes Attr;
    ImlibImage *Image;
    
    D = Dsk;
    Command = strdup( Com );
    Caption = strdup( Cap );
    this->X = X;
    this->Y = Y;
    Dragged = Dragging = False;
    strncpy(FileName, File, 128);
    FreeAlpha = false;
    Glowing = false;
    
    if( LoadPng( F ) == -1 ) {
      Image = Imlib_load_image( D->Data, F );
      if(! Image ) { printf("Couldn't find image %s, bailing.\n", F ); exit(-1); }
      rgb = Image->rgb_data;
      Width = Image->rgb_width;
      Height = Image->rgb_height;
      if (!(Image->alpha_data)) {
      	FreeAlpha = true;
	unsigned char trans = 255 - (unsigned char)D->GetTransparency();
	// Would rather use new/delete but not sure how that would behave with free() on non-gcc systems.
	// Also, would it be better to put this in a separate function? -- csm
	alpha = (unsigned char *)malloc(Width * Height * sizeof(unsigned char));
	alpha2 = (unsigned char *)malloc(Width * Height * sizeof(unsigned char));
        memset(alpha, trans, Width * Height * sizeof(unsigned char));
        memset(alpha2, 255, Width * Height * sizeof(unsigned char));
      } else {
        alpha = Image->alpha_data;
	alpha2 = Image->alpha_data;
      }
    }
    Attr.override_redirect = True;
    Attr.event_mask = SubstructureRedirectMask|SubstructureNotifyMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask|EnterWindowMask|LeaveWindowMask;
    Win  = XCreateWindow( D->Dpy, D->Root, 0, 0, Width, Height, 0, CopyFromParent, CopyFromParent, CopyFromParent, CWOverrideRedirect|CWEventMask, &Attr );
    XLowerWindow( D->Dpy, Win  );
    
    Attr.event_mask = SubstructureRedirectMask|SubstructureNotifyMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask|ExposureMask;
    XftTextExtents8( D->Dpy, D->Font, (XftChar8*)Caption, strlen(Caption), &FontInfo );
    Text = XCreateWindow( D->Dpy, D->Root, 0, 0, FontInfo.width + 4, D->Font->height, 0, CopyFromParent, CopyFromParent, CopyFromParent, CWOverrideRedirect|CWEventMask, &Attr );
    Draw = XftDrawCreate( D->Dpy, Text, DefaultVisual(D->Dpy,DefaultScreen(D->Dpy)), DefaultColormap(D->Dpy,DefaultScreen(D->Dpy)) );
    XLowerWindow( D->Dpy, Text );  
  }


  Icon::~Icon( ) {
    free( Command );
    free( Caption );
    if( FreeAlpha ) {
      free( alpha );
      free( alpha2 );
    }
  }


  void Icon::RefreshIcon( ) {
      ImlibImage *Bg;
      unsigned char *bgrgb;
      unsigned char *alpha_levels;
      
      if( Glowing )
	alpha_levels = alpha2;
      else
	alpha_levels = alpha;

      Bg = Imlib_crop_and_clone_image( D->Data, D->SpareRoot, X, Y, Width, Height );
      bgrgb = Bg->rgb_data;
      
      for( int i=0, a=0; i<(3*Width*Height); i+=3, a++) {
         int alevel = ( (alpha_levels != 0) ? alpha_levels[a] : 255 );
         if( alevel == 0 ) { bgrgb[i] = bgrgb[i+2] = 255; bgrgb[i+1] = 0; continue; }
         bgrgb[i+0] = (bgrgb[i+0] * (255-alevel))/255 + (rgb[i+0] * alevel)/255;
         bgrgb[i+1] = (bgrgb[i+1] * (255-alevel))/255 + (rgb[i+1] * alevel)/255;
         bgrgb[i+2] = (bgrgb[i+2] * (255-alevel))/255 + (rgb[i+2] * alevel)/255;
      }
      
      ImlibColor IClr={255,0,255,0};
      Imlib_set_image_shape(D->Data, Bg, &IClr);
      Imlib_apply_image( D->Data, Bg, Win );

      Imlib_kill_image( D->Data, Bg );
  }


  void Icon::Update( ) {
      ImlibImage *TextBg;
      int fX, fY;
       
      XUnmapWindow( D->Dpy, Win );
      XUnmapWindow( D->Dpy, Text );
      
      fY = Y + Height + 4;
      fX = X + ( Width - FontInfo.width )/2;
      TextBg = Imlib_create_image_from_drawable( D->Data, D->Root, 0, fX, fY, FontInfo.width + 4, D->Font->height );
      Imlib_apply_image( D->Data, TextBg, Text );
      XMoveWindow( D->Dpy, Text, fX, fY );

      RefreshIcon();
       
      XMapWindow( D->Dpy, Win );
      XMapWindow( D->Dpy, Text );
      
      UpdateText( );
  }


  void Icon::UpdateText( ) 
  {
    int ShadowX, ShadowY;

    XClearWindow( D->Dpy, Text );
    
    if( D->GetContrast() )
    {
      // Crude edging.  I would prefer this blurred.
      for( int i = 0; i < 5; i++ )    
      {
	XftDrawString8( Draw, &D->Back, D->Font, i, D->Font->ascent, (XftChar8*)Caption, strlen(Caption) );
	XftDrawString8( Draw, &D->Back, D->Font, 2, D->Font->ascent + i - 2, (XftChar8*)Caption, strlen(Caption) );
      }
    }
    
	if ( D->GetShadow(ShadowX, ShadowY) )
		XftDrawString8( Draw, &D->Shadow, D->Font, 2 + ShadowX, D->Font->ascent + ShadowY, (XftChar8*)Caption, strlen(Caption) );
    XftDrawString8( Draw, &D->Fore, D->Font, 2, D->Font->ascent, (XftChar8*)Caption, strlen(Caption) );
  }


  void Icon::TwiddleGlow( bool NewSetting ) {
    if( Glowing != NewSetting )
    {
      Glowing = NewSetting;
      RefreshIcon();
    }
  }


  void Icon::Drag( XEvent *Ev ) {
    static int lX = 0, lY = 0, Time = 0;
    int fX, fY;
    
    bool locked = D->IsLocked();

    if( Ev->type == ButtonRelease )
    {
      // Something of a special case.
      if( Ev->xbutton.time - Time <= 200 )
      {
	D->Execute( Command );
	Time = 0;
      }
      else Time = Ev->xbutton.time;
    }
     
    if( locked ) return;

    switch( Ev->type )
    {
    case ButtonPress:
      lX = Ev->xbutton.x_root;
      lY = Ev->xbutton.y_root;
      Dragging = True;
      break;
    case MotionNotify:
      Dragged = True;
      X  -= lX - Ev->xmotion.x_root;
      Y  -= lY - Ev->xmotion.y_root;
      lX  = Ev->xmotion.x_root;
      lY  = Ev->xmotion.y_root;
      if( X < 0 ) X = 0;
      if( Y < 0 ) Y = 0;
      if( X+Width > WidthOfScreen(DefaultScreenOfDisplay(D->Dpy)) ) X = WidthOfScreen(DefaultScreenOfDisplay(D->Dpy)) - Width;
      if( Y+Height > HeightOfScreen(DefaultScreenOfDisplay(D->Dpy)) ) Y = HeightOfScreen(DefaultScreenOfDisplay(D->Dpy)) - Height;
      fY = Y + Height + 4;
      fX = X + ( Width - FontInfo.width )/2;
      XMoveWindow( D->Dpy, Win, X, Y );
      XMoveWindow( D->Dpy, Text, fX, fY );
      break;
    case ButtonRelease:
      if( Dragged ) { Update( ); D->SaveIcon(this); }
      Dragging=False;
      Dragged=False;
      break;
    }
    
  }


 int Icon::LoadPng( char *filename ) {
   unsigned char *ptr, *aptr, *aptr2, **lines, *line, r, g, b, a;
   png_structp   pngdata;
   png_infop     pnginfo;
   png_uint_32   w, h;
   FILE          *file;
   int           color, depth, interlace;

     if(! (file = fopen( filename, "r" )) ) return -1;

     pngdata = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL );
     if(! pngdata ) return -1;
     pnginfo = png_create_info_struct( pngdata );
     if(! pnginfo ) { png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }

     if( setjmp(pngdata->jmpbuf) ) { png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }
     if( color == PNG_COLOR_TYPE_RGB_ALPHA ) { png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }

     png_init_io( pngdata, file );
     png_read_info( pngdata, pnginfo );
     png_get_IHDR( pngdata, pnginfo, &w, &h, &depth, &color, &interlace, NULL, NULL );
 
     Width = w;
     Height = h;

     if( color == PNG_COLOR_TYPE_PALETTE ) png_set_expand( pngdata );
     png_set_strip_16( pngdata );
     png_set_packing( pngdata );

     if( png_get_valid(pngdata,pnginfo,PNG_INFO_tRNS) ) png_set_expand( pngdata );
     png_set_filler( pngdata, 0xff, PNG_FILLER_AFTER );


     rgb = (unsigned char *)malloc( Width * Height * 3 );
     if(! rgb ) { png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }
     alpha = (unsigned char *)malloc( Width * Height );
     alpha2 = (unsigned char *)malloc( Width * Height );
     if( !alpha || !alpha2 ) { png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }
     FreeAlpha = true;

     lines = (unsigned char **)malloc(Height * sizeof(unsigned char *));
     if(! lines ) { delete(rgb); delete(alpha); delete(alpha2); png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }
     for(int i=0; i< Height; i++) {
       lines[i] = (unsigned char *)malloc( Width * (sizeof(unsigned char*)) * 4 );
       if(! lines[i] ) { delete(rgb); delete(alpha); delete(alpha2); png_destroy_read_struct( &pngdata, NULL, NULL ); return -1; }
     }

     png_read_image( pngdata, lines );
     png_destroy_read_struct( &pngdata, &pnginfo, NULL );
     ptr = rgb;
     aptr = alpha;
     aptr2 = alpha2;
     
     for(int y = 0; y < Height; y++) {
       line = lines[y];
       for (int x = 0; x < Width; x++) {
         if( color == PNG_COLOR_TYPE_GRAY_ALPHA ) {
           r = g = b = *line++;
           a = *line++;
         } else if( color == PNG_COLOR_TYPE_GRAY ) {
           r = g = b = *line++;
           a = 255;
         } else {
           r = *line++;
           g = *line++;
           b = *line++;
           a = *line++;
         }
          *ptr++ = r;
          *ptr++ = g;
          *ptr++ = b;
	  
	  // Transparency matrix.
	  if( ( a - D->GetTransparency() ) < 0 )
	  {
	    *aptr++ = 0;
	    *aptr2++ = a;
	  }
	  else 
	  {
	    *aptr++ = a - D->GetTransparency();
	    *aptr2++ = a;
	  }
	  
       }
     } 

   for(int i=0; i<Height; i++) free(lines[i]);
   free(lines);
   return 0;
 }
